/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

#include "eclib.h"

extern int errno;

char ZC_alev;
extern short ZC_max_struct, ZC_max_malloc;

/* The following staticvariables start off 0, thanks to crt0 */
extern struct {
char *block;
unsigned char level;
} ZC_inmem[];
static unsigned short num_malloc; /* number of blocks allocated */
/* for each level, the current block and position */
static char *cur_block[MAX_ALEV];
static unsigned short cur_pos[MAX_ALEV];
/*  a chain of available nodes for each type of structure and each level */
extern char *struct_avail[1][MAX_ALEV];
/* # of gets, frees, and size of avail list for each structure and level */
extern long num_gets[][MAX_ALEV],
	num_frees[][MAX_ALEV],
	num_avails[][MAX_ALEV];

zap_alloc()
{
register i, j;

for(i=0; i<ZC_max_struct; ++i) {
num_avails[i][ZC_alev] = num_frees[i][ZC_alev] = num_gets[i][ZC_alev] = 0;
struct_avail[i][ZC_alev] = 0;
}
cur_block[ZC_alev] = 0;
cur_pos[ZC_alev] = 0;

for(i=j=0; i<num_malloc; ++i) {
if(ZC_inmem[i].level == ZC_alev) { /* drop this one */
free(ZC_inmem[i].block);
continue;
}
ZC_inmem[j] = ZC_inmem[i]; /* structure copy */
++j;
} /* end loop dropping blocks at the given level */
num_malloc = j;
} /* zap_alloc */

static char *grab_space(n)
unsigned int n;
{
    char *s, *malloc();
    
        if (n > BYTES_MALLOC)
eclib_error(NODE_ALLOC_SIZE, n);
    
    if (!cur_block[ZC_alev] ||
/* beware of the following line.  MSDOS integer overflow could be a problem.
 * With n below 32K and BYTES_MALLOC at 32K, the sum will not wrap around.
 * Don't increase BYTES_MALLOC beyond 32K, or your 16-bit word might overflow,
 * this test will pass when it should not, and you'll be in trouble.
 * Furthermore, I'm not sure the Microsoft compiler can access s[40000].
 * You might wind up with s[-25536].  So stay below 32K. */
cur_pos[ZC_alev] + n > BYTES_MALLOC) {

if(num_malloc >= ZC_max_malloc) {
eclib_error(NODE_NOMEM_TOTAL, ZC_max_malloc);
return 0;
}

errno = 0;
            s = malloc( (unsigned) BYTES_MALLOC);
            if (!s) eclib_error(NODE_ALLOC_FAILURE, errno);

cur_block[ZC_alev] = ZC_inmem[num_malloc].block = s;
ZC_inmem[num_malloc].level = ZC_alev;
++num_malloc;
cur_pos[ZC_alev] = 0;
	}

s = cur_block[ZC_alev] + cur_pos[ZC_alev];
    cur_pos[ZC_alev] += n;
    return s;
}  /* grab_space */

char **ZC_serial_grabblock(n, struct_type)
unsigned int n;
{
char *s, *malloc();

if(num_malloc >= ZC_max_malloc) {
eclib_error(NODE_NOMEM_TOTAL, ZC_max_malloc);
return 0;
}

if(ZC_alev < struct_desc[struct_type].minlev || ZC_alev > struct_desc[struct_type].maxlev)
eclib_error(NODE_LEVEL_STRUCT_MISMATCH, struct_type);

errno = 0;
            s = malloc(n);
            if (!s) eclib_error(NODE_ALLOC_FAILURE, errno);

ZC_inmem[num_malloc].block = s;
ZC_inmem[num_malloc].level = ZC_alev;
return &ZC_inmem[num_malloc++].block;
} /* ZC_serial_grabblock */

char *ZC_get_struct(struct_type)
int struct_type; /* what kind of structure to allocate */
{
char *s;
int i;
long space;
    
if(struct_desc[struct_type].nodetype == 1)
eclib_error(NODE_STRUCT_TYPE, struct_type);

if(ZC_alev < struct_desc[struct_type].minlev || ZC_alev > struct_desc[struct_type].maxlev)
eclib_error(NODE_LEVEL_STRUCT_MISMATCH, struct_type);

space = 0;
for(i=0; i<MAX_ALEV; ++i)
space += num_gets[struct_type][i] - num_frees[struct_type][i];
if(struct_desc[struct_type].memlimit > 0) {
space *= struct_desc[struct_type].size;
space /= 1024;
if(space >= struct_desc[struct_type].memlimit) {
eclib_error(NODE_NOMEM_STRUCT, struct_type);
return 0;
}
}

++num_gets[struct_type][ZC_alev];
    if (struct_avail[struct_type][ZC_alev] == 0)
        s =  grab_space(struct_desc[struct_type].size);
    else {
--num_avails[struct_type][ZC_alev];
s = struct_avail[struct_type][ZC_alev];
struct_avail[struct_type][ZC_alev] = * (char **)s;
if(((long *)s)[1] != AVAIL_CONSTANT)
eclib_error(NODE_CORRUPT, struct_type);
        }

/* init the structure to initchar. */
if(struct_desc[struct_type].initchar >= 0)
memset(s, struct_desc[struct_type].initchar, struct_desc[struct_type].size);
    return s;
}  /* ZC_get_struct */

ZC_free_struct(s, struct_type)
char *s; /* pointer to structure to free */
int struct_type; /* type of structure to free */
{
if(struct_desc[struct_type].nodetype == 1)
eclib_error(NODE_STRUCT_TYPE, struct_type);

if(s == 0)
eclib_error(NODE_NULL_PTR, 0);
++num_frees[struct_type][ZC_alev];
++num_avails[struct_type][ZC_alev];
((long *)s)[1] = AVAIL_CONSTANT;
* (char **)s = struct_avail[struct_type][ZC_alev];
struct_avail[struct_type][ZC_alev] = s;
}  /* ZC_free_struct */

char *ZC_copy_struct(s, struct_type)
char *s; /* node being copied */
int struct_type; /* what kind of structure to allocate */
{
char *t = ZC_get_struct(struct_type);
if(struct_desc[struct_type].nodetype < 0)
eclib_error(NODE_STRUCT_TYPE, struct_type);
memcpy(t, s, struct_desc[struct_type].size);
return t;
} /* ZC_copy_struct */

set_alev(level)
unsigned char level;
{
if(level > MAX_ALEV) eclib_error(NODE_BAD_LEVEL, level);
else ZC_alev = level;
} /* set_alev */

get_alev()
{
return ZC_alev;
} /* get_alev */

print_mem()
{
int i, j;

    printf("------------- memory usage ------------\n");
    printf("%d mallocs consuming %dK\n",
num_malloc, num_malloc*K_MALLOC);

printf("struct[level]: gets frees avails memory-consumed\n");
for(i=0; i<ZC_max_struct; ++i)
for(j=0; j<MAX_ALEV; ++j)
if(num_gets[i][j] != num_frees[i][j])
  printf("%s[%d]:   %ld %ld %ld %ldK\n",
struct_desc[i].name, j, num_gets[i][j], num_frees[i][j],
num_avails[i][j],
((num_gets[i][j]-num_frees[i][j]) * struct_desc[i].size) / 1024);
printf("\n");
}  /* print_mem */

check_mem_empty()
{
register i;

for(i=0; i<ZC_max_struct; ++i)
if(num_gets[i][ZC_alev] != num_frees[i][ZC_alev]) {
print_mem();
eclib_error(NODE_NONEMPTY, i);
}
} /* check_mem_empty */

ec_zap_chain(s, struct_type)
char *s;
int struct_type; /* kind of structure */
{
char *t;

while(s) {
t = s;
s = *((char **)s); /* s = s->next */
ZC_free_struct(t, struct_type);
}
} /* ec_zap_chain */

