/***********************************************************************
|  Enhanced C library
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

#define YYFLAG		(-1000)
# define YYERRCODE 1
#define MAX_PARSE_STACK 150

char ZC_parsedeb; /* parser debugging */

extern char ZC_recflag; /* error recovery flag */
extern short ZC_rules[], ZC_rule_offset[], ZC_nt_base;
extern int ZC_curr_token; /* current input token number */

/* this union is actually defined by the user's program.
 * may be larger than the one shown here. */
typedef union { long a; /* at least 4 bytes */ } ZC_gramunion;
extern short ZC_gramunion_size;
extern ZC_gramunion ZC_inp_parseval, ZC_curr_parseval;


/* parser, 0 return means success, 1 means an error */
parse_go()
{
		char *v_ptr;	/* top of value stack */
		short *s_ptr;		/* top of state stack */
		short state;		/* current state */
		short  n, save_n; /* state, rule number, or lhs */
	short len; /* length of production */
int rc; /* return code indicating accept, abort, or continue */

static char *values; /* allocated stack, ZC_gramunion[MAX_PARSE_STACK] */
static short states[ MAX_PARSE_STACK ];		/* state stack */
extern char *emalloc();

extern short ZC_yylast;
extern 	char *ZC_stacktop;	/* top of value stack for $vars */
/* generated tables that drive the parser */
extern short ZC_exca[];
extern short ZC_act[];
extern short ZC_pact[];
extern short ZC_pgo[];
extern short ZC_prod_length[];
extern short ZC_chk[];
extern short ZC_def[];

if(!values) /* allocate stack */
values = emalloc(MAX_PARSE_STACK * ZC_gramunion_size);

	ZC_recflag = 0;
	ZC_curr_token = -1;

	v_ptr = values;
	s_ptr = &states[-1];
	state = 0;

	yystack:
		/* push state and value onto the stack */
		if ( ++s_ptr >= &states[ MAX_PARSE_STACK ] ) {
			parse_error( "parser stack overflow" );
return 1;
		}
		*s_ptr = state;
		memcpy(v_ptr, &ZC_curr_parseval, ZC_gramunion_size);
		v_ptr += ZC_gramunion_size;

		/* new state, determine appropriate action */
	newstate:
		if ( ( n = ZC_pact[ state ] ) <= YYFLAG )
			goto simple_act;		/* simple state */

		token_pending();

		if ( ( ( n += ZC_curr_token ) < 0 ) || ( n >= ZC_yylast ) )
			goto simple_act;
		if ( ZC_chk[ n = ZC_act[ n ] ] == ZC_curr_token )	/*valid shift*/
		{
			ZC_curr_token = -1;
			memcpy(&ZC_curr_parseval, &ZC_inp_parseval, ZC_gramunion_size);
			state = n;
			if ( ZC_recflag > 0 )
				ZC_recflag--;
			goto yystack;
		}

	simple_act:
		if ( ( n = ZC_def[ state ] ) == -2 ) {

			token_pending();

			/* look through exception table */
			{
				register short *yyxi = ZC_exca;

				while ( ( *yyxi != -1 ) ||
					( yyxi[1] != state ) )
					yyxi += 2;
				while ( ( *(yyxi += 2) >= 0 ) &&
					( *yyxi != ZC_curr_token ) )
					;
				if ( ( n = yyxi[1] ) < 0 )
return 0;
			}
		}

		/*
		** check for syntax error
		*/
		if ( n == 0 )	/* have an error */
		{
			/* no worry about speed here! */
			switch ( ZC_recflag )
			{
			case 0:		/* new error */
				parse_error( "syntax error" );
			case 1:
			case 2:		/* incompletely recovered error */
					/* try again... */
				ZC_recflag = 3;
				/* find state where "error" is a legal
				** shift action. */
				while ( s_ptr >= states )
				{
					n = ZC_pact[ *s_ptr ] + YYERRCODE;
					if ( n >= 0 && n < ZC_yylast &&
						ZC_chk[ZC_act[n]] == YYERRCODE)					{
						/*
						** simulate shift of "error"
						*/
						state = ZC_act[ n ];
						goto yystack;
					}
					/*
					** current state has no shift on
					** "error", pop stack
					*/
					s_ptr--;
					v_ptr -= ZC_gramunion_size;
				}
				/*
				** there is no state on stack with "error" as
				** a valid shift.  give up.
				*/
return 1;
			case 3:		/* no shift yet; eat a token */
				if ( ZC_curr_token == 0 )	/* reached EOF. quit */
return 1;
				ZC_curr_token = -1;
				goto newstate;
			}
		}/* end if ( n == 0 ) */
		/* apply production n */
		if(ZC_parsedeb)
show_prod(n);

		ZC_stacktop = v_ptr;			/* $vars top of value stack */
		save_n = n;			/* value to switch over */
len = ZC_prod_length[n] >> 1;

				/* default assignment of $$ */
				 v_ptr -= len*ZC_gramunion_size ;
				memcpy(&ZC_curr_parseval, v_ptr, ZC_gramunion_size);
				state = ZC_pgo[ n = ZC_rules[ZC_rule_offset[n]] - ZC_nt_base ] +
					*( s_ptr -= len ) + 1;
				if ( state >= ZC_yylast ||
					ZC_chk[ state = ZC_act[ state ] ] != -n )
					state = ZC_act[ ZC_pgo[ n ] ];
				if(!(ZC_prod_length[save_n] & 1)) goto yystack;

rc = parser_rules(save_n);
if(rc < 0) goto yystack;
return rc;
} /* parse_go */

static char *show_symbol(sn)
short sn; /* symbol number */
{
extern char *ZC_t_names[], *ZC_nt_names[]; /* names for gramatical units */
if(sn < ZC_nt_base) return ZC_t_names[sn];
return ZC_nt_names[sn-ZC_nt_base];
} /* show_symbol */

show_prod(n)
{
short start, end;
extern char *show_symbol();

start = ZC_rule_offset[n];
end = ZC_rule_offset[n+1];
printf("reduction %s ->", show_symbol(ZC_rules[start]));
for(++start; start < end; ++start)
printf(" %s", show_symbol(ZC_rules[start]));
printf("\n");
} /* show_prod */

static token_pending()
{
extern char *show_symbol();
if(ZC_curr_token < 0) {
if((ZC_curr_token = lex_go()) < 0)
ZC_curr_token = 0;
if(ZC_parsedeb)
printf("token %s\n", show_symbol(ZC_curr_token));
}
} /* token_pending */

