/***********************************************************************
|  Interactively configurable forms
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* grammar.ec: grammar and scanner for icf. */

/* The language is a sequence of statementss,
 * where each statement establishes a domain,
 * defines a question, etc. */

${
#include <setjmp.h>

#define EXP_VARNUM 20
#define EXP_OP 21
#define EXP_BUILTIN 22
#define EXPLEN 30 /* largest expression */
#define ARRAYLEN 20 /* length of array of answers */
#define STACKSIZE 400
#define ARRAYBIT 0x1000 /* bit indicating array */
#define is_array(b) ((b)->stringmod&0x10)
#define blockname(b) code_to_string((b)->name)

#define BLK_QUES 1 /* block is a question */
#define BLK_DECL 2 /* block is a declaration */
#define BLK_IF 3 /* block is an if jump */
#define BLK_WHILE 4 /* block is an while loop */
#define BLK_ASG 5 /* variable assignment */
#define BLK_ARPOP 6 /* pop array */
#define BLK_CONT 7 /* continue */
#define BLK_BREAK 8 /* break */
#define BLK_ECHO 9
#define BLK_LA 10 /* loop over array */
#define BLK_ARCLEAR 11
#define BLK_ENDARG 12
#define BLK_RET 13
#define BLK_CALL 14
#define BLK_PARMSET 15
#define BLK_ENDFN 16

#define TYPE_MONEY 1
#define TYPE_INT 2
#define TYPE_FN 3
#define TYPE_STRING 4
#define TYPE_BOOL 5
#define TYPE_DATE 6

extern jmp_buf actjmp;
extern int icfdebug, inbatch;
extern struct serial_inmem blist;
extern char streamuse[100], *tfname();
extern FILE *formfile, *ansfile;
extern char *ansfilename;
extern struct block *bnum_to_block(), *bname_to_block();

struct expstr {
long *e;
int size_e;
};

$}

global:
usage "usage:  icf [-v] [-f] [-l length] [-w width] formdescfile session_file":
eprint_exit exit;

main:
options "dvfl^w^":
arguments "..";

node list order_by "uu" domain_member { /* a domain element */
short elnum; /* element number */
#define cmp_domain_member(s,t) (s->elnum-t->elnum)
str_code elsym; /* element symbol */
#define cmp_domain_member_1(s,t) ci_strncmp(code_to_string(s->elsym), code_to_string(t->elsym), 0)
};

node list order_by "uu" domain {
struct list_domain_member members;
short dnum;
#define cmp_domain(s,t) (s->dnum-t->dnum)
str_code name;
#define cmp_domain_1(s,t) ((int)s->name-(int)t->name)
str_code helpmsg;
}: initchar 0: gvar domains;

/* thanks to the code in bname_to_block(), this structure
 * needs to be a multiple of 4 bytes, in small or large model */
node serial block {
short blktype; /* question, declaration, if test, etc */
short bnum;
str_code name;
char vartype, stringmod;
long ask; /* question to be asked */
long nextask; /* second question to be asked */
long explain; /* explanation, if you don't understand the question */
long answer; /* answer to the question*/
long *formula; /* if formula then jump to line 30 etc */
short fmsize; /* length of formula */
/* the above array of longs is also used to hold a sequence of
 * answers, when the variable is declared as an array. */
short gotonum; /* if formula goto gotonum */
char notnull; /* has question been answered */
char fromstack;
short filler;
};


scanner:
string_length 200:
string_desc "dcn":
string_storage:

token_strings:

/* a word, which may identify a domain or a question etc,
 * is at most 40 characters, which really ought to be enough. */
word_length 40:
word_storage:
word_fn identmatch:

/* shell style comments */
divert "/* */":

/* build C style formulas */
left "=" ASG ASG:
newprec:
left "or" OR OR:
newprec:
left "and" AND AND:
newprec:
prec "not" NOT NOT:
newprec:
left "==" EQ RELOP:
left "!=" NE RELOP:
left "<" LT RELOP:
left "<=" LE RELOP:
left ">" GT RELOP:
left ">=" GE RELOP:
newprec:
left "|" PIPE PIPE:
newprec:
left "+" ADD ADDOP:
left "-" SUB ADDOP:
newprec:
left "*" MUL MULOP:
left "/" DIV MULOP:
left "%" REM MULOP:
newprec:
left "using" USING USING:

/* helpful instructions or canned paragraphs */
token "`" TEXT TEXT 0 skip_to_etext:

/* types of entered data */
token "money" MONEY TYPE TYPE_MONEY:
token "int" INT TYPE TYPE_INT:
token "string" STRING TYPE TYPE_STRING:
token "raw" RAW STRINGMOD 0:
token "downshift" DOWNSHIFT STRINGMOD 1:
token "pretty" PRETTY STRINGMOD 2:
token "upshift" UPSHIFT STRINGMOD 3:
token "bool" BOOL TYPE TYPE_BOOL:
token "date" DATE TYPE TYPE_DATE:

token "false" FALSE BOOLCONST 0:
token "true" TRUE BOOLCONST 1:
token "$" MONEYCONST MONEYCONST 0 parsemoney:

/* simple punctuations */
token "," COMMA COMMA:
token ":" COLON COLON:
token ";" SEMI SEMI:
token "{" LC LC:
token "}" RC RC:
token "(" LP LP:
token ")" RP RP:
token "[" LB LB:
token "]" RB RB:

/* other types of tokens, reflecting nodes above */
my_type_ptr domain:
my_type_ptr block:
my_type expstr:

/* aggrigate operations on arrays */
token "array" ARRAY ARRAY:
token "max" MAX AGGR:
token "min" MIN AGGR:
token "length" LENGTH AGGR:
token "sum" SUM AGGR:
token "avg" AVG AGGR:
token "pop" POP POP 6:
token "popall" POPALL POP 11:

/* branching and control */
right "if" IF COND BLK_IF:
right "while" WHILE COND BLK_WHILE:
right "else" ELSE ELSE BLK_IF:
token "looparray" LOOPARRAY LOOPARRAY:
token "continue" CONT CONT BLK_CONT:
token "break" BREAK CONT BLK_BREAK:

/* other reserved words */
token "domain" DOMAIN DOMAIN:
token "echo" ECHO ECHO BLK_ECHO:
token "decl" DECLARE DECLARE BLK_DECL:
token "ask" ASK ASK:
token "nextask" NEXTASK NEXTASK:
token "call" CALL CALL:
token "function" FUNCTION FUNCTION:
token "returning" RETURNING RETURNING:
token "return" RETURN RETURN:
token "explain" EXPLAIN EXPLAIN:
token "prompt" QUESTION QUESTION BLK_QUES;


parser:

cast_str_code WORD STRING call2:
cast_long NUMBER MONEYCONST instruct TEXT:
cast_tok_extend STRINGMOD TYPE rescond COND ELSE loopcond CONT:
cast_tok_extend QUESTION DECLARE qdecl array BOOLCONST POP ECHO:
cast block vtype retlist parm fn1:
cast domain domlist:
cast expstr exp arlist:
cast_tok_val AND OR NOT RELOP MULOP ADDOP PIPE AGGR USING:

rule[input]
|	/* empty */
|	input stmt
|	input fn2 stmt = simple_stmt(BLK_ENDFN, cast_str_code 0)
:

rule[fn2]
|	fn1 LP parmlist RP = endargs(0)
:

rule[parmlist]
|	parm
|	parmlist COMMA parm
:

rule[retlist]
|	WORD < make_vtype(BLK_ASG, 0, cast_str_code 0, 0, 0, $1)
|	retlist COMMA WORD < make_vtype(BLK_ASG, 0, cast_str_code 0, 0, 0, $3)
:

rule[parm]
|	TYPE WORD < make_vtype(BLK_ASG,0,cast_str_code 0, $1, 2, $2)
|	TYPE STRINGMOD WORD < make_vtype(BLK_ASG,0,cast_str_code 0, $1, $2,$3)
|	WORD WORD < make_vtype(BLK_ASG,0,$1,0,0,$2)
:

rule[fn1]
|	FUNCTION WORD < make_vtype(BLK_DECL, 0, cast_str_code 0, TYPE_FN, 0, $2)
:

rule[stmtlist] /* list of blocks */
|	stmt
|	stmtlist stmt
:

rule[domlist] /* define a domain */
|	STRING < setdom_mm(cast domain 0, $1) /* first domain member */
|	domlist COMMA STRING < setdom_mm($1,$3)
:

rule[arlist] /* list of arrays */
|	WORD < mkarlist(cast expstr 0, $1)
|	arlist COMMA WORD < mkarlist($1,$3)
:

rule[arglist]
|	exp = pusharg($1)
|	exp COMMA arglist = pusharg($1)
:

rule [rescond] /* reserve a block for conditional jump */
|	< cond_reserve_block()
:

rule[loopcond]
|	LOOPARRAY arlist < start_looparray($2)
:

rule[call1]
|	call2 LP arglist RP = callfunc($1)
:

rule[call2]
|	CALL WORD < frameset($2)
:

rule[stmt]
|	SEMI /* empty statement */
|	LC stmtlist RC /* block of statements, C style */
|	DOMAIN WORD STRING LC domlist RC SEMI
	=setdom($2,$3,$5)
|	vtype ASK TEXT instruct SEMI
	=add_varblock($1,$3, cast_long 0,$4)
|	vtype ASK TEXT NEXTASK TEXT instruct SEMI
	=add_varblock($1,$3,$5,$6)
|	vtype SEMI
	=add_varblock($1,cast_long 0,cast_long 0, cast_long 0)
|	vtype ASG exp SEMI
	= add_varasg($1,$3)
|	WORD ASG exp SEMI = varassign($1,$3)
|	POP WORD SEMI = simple_stmt($1,$2)
|	CONT SEMI = simple_stmt($1,cast_str_code 0)
|	call1 SEMI = endargs(1)
|	call1 RETURNING retlist SEMI = endargs(1)
|	RETURN arglist SEMI = simple_stmt(BLK_RET, cast_str_code 0)
|	RETURN SEMI = simple_stmt(BLK_RET, cast_str_code 0)
|	ECHO TEXT SEMI = echo_text(cast_long 0,$2)
|	ECHO NUMBER TEXT SEMI = echo_text($2,$3)
|	COND rescond exp stmt
	= add_condition($3,$2,$1,0)
|	COND rescond exp stmt ELSE rescond stmt
	= add_condition($3,$2,$1,$6)
|	loopcond stmt
	= add_condition(cast expstr 0, $1, BLK_LA, 0)
:

/* instructions are optional */
rule[instruct]
|	/* empty */ < cast_long 0
|	EXPLAIN TEXT <$2
:

rule[qdecl]
|	QUESTION <$1
|	DECLARE <$1
:

rule[array]
|	/* empty */ <0
|	ARRAY <1
:

rule[vtype] /* variable and its type */
|qdecl array TYPE WORD < make_vtype($1,$2,cast_str_code 0, $3, 2, $4)
|qdecl array TYPE STRINGMOD WORD < make_vtype($1,$2,cast_str_code 0, $3, $4,$5)
|qdecl array WORD WORD < make_vtype($1,$2,$3,0,0,$4)
|qdecl array WORD < make_vtype($1,$2,cast_str_code 0,0,0,$3)
:

rule[exp] /* expression */
|	exp OR exp < buildexp($1,$2,$3)
|	exp AND exp < buildexp($1,$2,$3)
|	NOT exp < monoexp($2, $1)
|	exp RELOP exp < buildexp($1,$2,$3)
|	exp MULOP exp < buildexp($1,$2,$3)
|	exp PIPE exp < buildexp($1,$2,$3)
|	exp ADDOP exp < buildexp($1,$2,$3)
|	ADDOP exp < monoexp($2,$1)
|	exp USING STRING < usingexp($1,$2,$3)
|	LP exp RP <$2
/*	WORD LB exp RB array indexing not yet implemented */
|	AGGR LP WORD RP < builtin($1,$3)
|	WORD < wordexp($1)
|	NUMBER < numexp($1, TYPE_INT)
|	MONEYCONST < numexp($1, TYPE_MONEY)
|	BOOLCONST < shortnumexp($1)
|	STRING < scnumexp($1)
;

${
/* globals corresponding to the externs above */
int icfdebug = 0, inbatch;
struct serial_inmem blist;
jmp_buf actjmp;
char streamuse[100];
FILE *formfile, *ansfile;
char *ansfilename;
$}
