/***********************************************************************
|  Interactively configurable forms
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* icfgsup.c: support routines for icfgram.ec */

#include "icfgram.h"

extern FILE *ZC_file;

static int in_looparray;

static struct block *newblock(blktype, vartype, blkname)
str_code blkname;
{
static struct block q;

q.bnum = blist.n_nodes;
q.blktype = blktype;
q.vartype = vartype; /* cuts off the bottom byte */
q.formula = 0;
if(vartype & ARRAYBIT)
q.formula = (long *) emalloc(4*ARRAYLEN);
q.stringmod = vartype>>8;
if(!vartype) blkname = 0;
q.name = blkname;
if(blkname) bname_to_block(blkname, 1, 0);
insupd_mem_block(&blist, 0L, &q);
return bnum_to_block(q.bnum);
} /* newblock */

cond_reserve_block()
{
struct block *b = newblock(BLK_IF, 0, (str_code)0);
return b->bnum;
} /* cond_reserve_block */

parsemoney(gramval)
{
long l = 0;
int numdig = 0;
char c;

while((c = getc(ZC_file)) != EOF) {
if(!isdigit(c)) break;
++numdig;
l = 10*l+c-'0';
}
l *= 100;
if(c == '.') { /* pennies */
c = getc(ZC_file);
if(isdigit(c)) {
l += 10*(c-'0');
c = getc(ZC_file);
if(isdigit(c)) {
l += c-'0';
c = getc(ZC_file);
}
}
} /* dot found */
if(isdigit(c) || numdig == 0 || numdig > 7)
eprint("improperly formatted money constant");
ungetc(c, ZC_file);
ZC_inp_parseval.l = l;
return gramval;
} /* parsemoney */

struct block *bnum_to_block(bnum)
{
return select_mem_block(&blist, (long)bnum);
} /* bnum_to_block */

struct block *bname_to_block(bname,isdef, cbn)
str_code bname;
int isdef; /* is this a variable definition */
short cbn; /* current block number */
{
int n = blist.n_nodes;
int i;
struct block *b, *b_g, *b_l, *b_def;
char *varname = code_to_string(bname);
char infunc, inmyfunc, isinfunc;

b_l = b_g = 0;
infunc = isinfunc =0;

if(n > 1) {
b = select_mem_block(&blist, 1L);
for(i=1; i<n; ++i,++b) {
if(b->name == bname) {
if(!infunc) b_g = b;
else b_def = b;
}
if(b->vartype == TYPE_FN) {
infunc = 1;
inmyfunc = 0;
b_def = 0;
}
if(infunc && b->bnum == cbn)
inmyfunc = 1;
isinfunc = 1;
if(b->blktype == BLK_ENDFN) {
infunc = 0;
if(inmyfunc && b_def) b_l = b_def;
}
} /* loop over blocks */

if(infunc) {
if(b_def && cbn == 0)
b_l = b_def;
isinfunc = 1;
}
} /* blocks present */

if(isdef&1) { /* defining this variable */
if(b_l || b_g && !isinfunc)
eprint("Variable %s is defined more than once", varname);
return 0;
}

/* reference here */
b = b_l;
if(!b) b = b_g;
if(!b) eprint("variable %s not defined", varname);

if(isdef&2  && b->vartype != TYPE_FN)
eprint("cannot call %s(), since %s is not a function", varname, varname);
if(!(isdef&2) && b->vartype == TYPE_FN)
eprint("cannot use function %s in an expression", varname);

return b;
} /* bname_to_block */

/* add a new symbolic constant to a domain */
struct domain *setdom_mm(list, newsym)
struct domain *list;
str_code newsym;
{
struct domain_member newmem;

if(!list) list = get_domain();
newmem.elsym = newsym;
newmem.elnum =list->members.n_nodes + 1;
copy_insert_domain_member(&(list->members), &newmem);
return list;
} /* setdom_mm */

/* create a new domain */
setdom(domname, helpmsg, list)
str_code domname, helpmsg;
struct domain *list;
{
list->name = domname;
list->helpmsg = helpmsg;
list->dnum = domains.n_nodes + 100;
insert_domain(&domains, list);
} /* setdom */

struct block *make_vtype(blktype, arraybit, domcode, basetype, stringmod, bname)
str_code domcode, bname;
{
struct domain d, *t;
char *varname = code_to_string(bname);
short vartype = basetype | (stringmod << 8);
struct block *q, *qi;

if(domcode) {
d.name = domcode;
t = find_domain_1(&domains, &d);
if(!t) eprint("%s is not a valid datatype or established domain", code_to_string(domcode));
vartype = t->dnum;
}

if(arraybit) {
if(!vartype)
eprint("cannot use keyword array without a type (e.g. array int or array date)");
vartype |= ARRAYBIT;
}

qi = 0;
if(!vartype) {
if(blktype == BLK_DECL)
eprint("%s must be declared using a type (e.g. date or int)", varname);
qi = bname_to_block(bname,0, 0);
}

q = newblock(blktype, vartype, bname);

q->gotonum = q->bnum;
if(qi) q->gotonum = qi->bnum;
if(blktype == BLK_ASG)
q->fromstack = 2;
return q;
} /* make_vtype */

/* equal if either is a substring of the other, ignoring case */
substrcmp(elsym1, elsym2)
str_code elsym1, elsym2;
{
char *sym1, *sym2;
char c1, c2;

sym1 = code_to_string(elsym1);
sym2 = code_to_string(elsym2);
while(*sym1 && *sym2) {
c1 = *sym1 | 0x20;
c2 = *sym2 | 0x20;
c1 -= c2;
if(c1) break;
++sym1, ++sym2;
}
return c1;
} /* substrcmp */

flatmain(arg1, arg2)
char *arg1, *arg2;
{
extern int scanner_base;
extern unsigned short ss_room;

if(opt_v) {
printf("icf version 1.0\n");
exit(0);
}

if(opt_d) icfdebug = ZC_parsedeb = 1;

/* Don't really want this, do we...
cd_basedir("icf");
*/
scanner_base = 10;
ss_room = 4000;
serial_setup(&blist, STRUCT_block, 0L, 500);

pushstream(arg1);
parse_go();

/* reopen the form file, so we can copy the canned paragraphs therein */
formfile = efopen(arg1, "r");
/* open the session file, containing all questions and answers */
ansfilename = arg2;
if(access(ansfilename, 0)) {
ansfile = efopen(ansfilename, "w");
fclose(ansfile);
}
ansfile = efopen(ansfilename, "r+");
inbatch = 1;
nextbatch();

action();
} /* flatmain */

parse_error(s)
char *s;
{
eprint("%s at symbol %s",
s, ZC_t_names[ZC_curr_token]);
} /* parse_error */

identmatch(identnum)
{
return identnum;
} /* word_fn */

add_varblock(q,ask1, ask2, ins_offset)
struct block *q;
long ask1, ask2, ins_offset;
{
char *varname;
struct block *qi;

qi = q;
if(q->gotonum)
qi = bnum_to_block(q->gotonum);
varname = blockname(qi);

if(q->blktype == BLK_DECL) {
if(ask1) eprint("cannot use question strings in a decl statement, use prompt instead");
return;
}

q->ask=ask1;
if(ask2) {
if(!is_array(qi))
eprint("a second question can only be used when %s is an array", varname);
q->nextask=ask2;
}
q->explain = ins_offset;
} /* add_varblock */

add_varasg(b, exp)
struct block *b;
struct expstr exp;
{
b->blktype = BLK_ASG;
b->formula = exp.e;
b->fmsize = exp.size_e;
} /* add_varasg */

add_condition(exp, ifnum, condtype, elsenum)
struct expstr exp;
{
struct block *b1, *b2;

b1 = bnum_to_block(ifnum);
b1->blktype = condtype;
b2 = 0;
if(elsenum) b2 = bnum_to_block(elsenum);
/* put in formula, if present */
if(exp.size_e) {
b1->formula = exp.e;
b1->fmsize = exp.size_e;
}

if(b2) { /* else clause present */
if(condtype != BLK_IF)
eprint("cannot use else clause unless it follows an if clause");
b1->gotonum = elsenum+1;
b2->gotonum = blist.n_nodes;
return;
}

if(condtype == BLK_WHILE || condtype == BLK_LA) {
b2 = newblock(BLK_IF, 0, (str_code)0);
b2->gotonum = ifnum;
cont_break(b1);
if(condtype == BLK_LA) --in_looparray;
}
b1->gotonum = blist.n_nodes;
} /* add_condition */

skip_to_etext(gramval)
{
int c;
long l = ftell(ZC_file);
while((c = getc(ZC_file)) != EOF) {
if(c == '\n') ++ZC_lineno;
if(c == '`') break;
}
ZC_inp_parseval.l = l;
if(c != '`')
eprint("EOF encountered before the closing `");
return gramval;
} /* skip_to_etext */

struct expstr getnewexp(size)
int size;
{
struct expstr x;
x.size_e = size;
x.e = (long*) emalloc(8*size);
return x;
} /* getnewexp */

/* word acts as variable, referring to an answer to a question */
struct expstr wordexp(wc)
str_code wc;
{
bname_to_block(wc,0, 0); /* just to make sure it's defined */
return numexp((long)wc, EXP_VARNUM);
} /* wordexp */

struct expstr numexp(l, numtype)
long l;
{
struct expstr x;
x = getnewexp(1);
x.e[0] = numtype;
x.e[1] = l;
return x;
} /* numexp */

struct expstr shortnumexp(val) { return numexp((long)val,5); }

struct expstr scnumexp(val) str_code val; { return numexp((long)val,4); }

struct expstr usingexp(exp, op, ustring)
struct expstr exp;
str_code ustring; /* using string */
{
return buildexp(exp, op, scnumexp(ustring));
} /* usingexp */

struct expstr builtin(op, wc)
str_code wc;
{
struct block *b = bname_to_block(wc,0, 0);
struct expstr x;

arraytest(b);
x = wordexp(wc);
x.e[0] = 4; /* string, don't expand this variable right away */
return monoexp(x, op);
} /* builtin */

struct expstr buildexp(x1, op, x2)
struct expstr x1, x2;
{
int size_e;
struct expstr x3;
long *lp;

size_e = x1.size_e + x2.size_e + 1;
if(size_e > EXPLEN)
eprint("formula too complex, limit %d terms and operators", EXPLEN);
x3 = getnewexp(size_e);
memcpy(x3.e, x1.e, x1.size_e*8);
memcpy(x3.e+x1.size_e*2, x2.e, x2.size_e*8);
lp = x3.e + x1.size_e*2 + x2.size_e*2;
lp[0] = EXP_OP;
lp[1] = op;
free(x1.e);
free(x2.e);
return x3;
} /* buildexp */

struct expstr monoexp(x, op)
struct expstr x;
{
return buildexp(numexp(0L, 2), op, x);
} /* monoexp */

varassign(wc, exp)
str_code wc;
struct expstr exp;
{
struct block *b = newblock(BLK_ASG, 0, (str_code)0);
b->gotonum = bname_to_block(wc,0, 0)->bnum;
b->formula = exp.e;
b->fmsize = exp.size_e;
} /* varassign */

arraytest(b)
struct block *b;
{
if(!is_array(b))
eprint("%s is not an array", blockname(b));
} /* arraytest */

simple_stmt(blktype, wc)
str_code wc;
{
struct block *b = newblock(blktype, 0, (str_code)0);
if(wc) {
struct block *b1 = bname_to_block(wc,0, 0);
arraytest(b1);
b->gotonum = b1->bnum;
}
} /* simple_stmt */

echo_text(streamnum, text)
long streamnum, text;
{
struct block *b = newblock(BLK_ECHO, 0, (str_code)0);
b->explain = text;
if(streamnum > 99)
eprint("invalid stream number, use [1-99]");
b->gotonum = streamnum;
} /* echo_text */

/* resolve break/continue jumps */
cont_break(b)
struct block *b;
{
short contnum = b->bnum;
short i = contnum;
short brknum = blist.n_nodes;

while(i < brknum) {
if(b->blktype == BLK_BREAK) {
b->gotonum = brknum;
b->blktype = BLK_IF;
}
if(b->blktype == BLK_CONT) {
b->gotonum = contnum;
b->blktype = BLK_IF;
}
++b, ++i;
}
} /* cont_break */

/* make a list of arrays */
struct expstr mkarlist(list, new)
struct expstr list;
str_code new;
{
struct block *b = bname_to_block(new,0, 0);
arraytest(b);
if(!list.size_e) {
list.e = (long*)emalloc(4*EXPLEN);
}
if(list.size_e == EXPLEN)
eprint("Too many arrays in list, limit %d", EXPLEN);
list.e[list.size_e++] = b->bnum;
return list;
} /* mkarlist */

start_looparray(exp)
struct expstr exp;
{
struct block *b = newblock(BLK_LA, 0, (str_code)0);
if(in_looparray)
eprint("nested loop arrays");
++in_looparray;
b->formula = exp.e; /* attach list of arrays */
b->fmsize = exp.size_e;
return b->bnum;
} /* start_looparray */

pusharg(exp)
struct expstr exp;
{
struct block *b = newblock(BLK_ASG, 0, (str_code)0);
b->fmsize = exp.size_e;
b->formula = exp.e;
b->fromstack = 1;
} /* pusharg */

endargs(fs)
{
struct block *b = newblock(BLK_ENDARG, 0, (str_code)0);
b->fromstack = fs;
} /* endargs */

callfunc(wc)
str_code wc;
{
struct block *b = newblock(BLK_CALL, 0, (str_code)0);
b->answer = wc;
} /* callfunc */

str_code frameset(wc)
str_code wc;
{
simple_stmt(BLK_PARMSET, (str_code)0);
return wc;
} /* frameset */

