/***********************************************************************
|  Basic calculator with polynomials
|  Copyright (C) Karl Dahlke 1994
|  This software may be freely used by anyone for any personal,
|  non-profit purpose, and may be freely distributed to others.
|*********************************************************************/

/* bcderiv.c: take derivatives and simplify rational expressions */

#include "bcgram.h"

/* take the derivative of a polynomial */
struct POLY *pderiv(a, v) /*###*/
struct POLY *a;
{
short i,j;
struct POLY *p;
struct TERM *s, *t;
short d_flag;

p = newpoly(a->nterms);
s = a->terms;
t = p->terms;

for(; s < a->terms + a->nterms; ++s) {
if(s->degree == 0) continue; /* constant */
t->degree = s->degree - 1;
d_flag = 0;

/* we assume polysimp has ben run, so each variable appears once */
for(j=i=0; i<s->nvars; ++i) {
if(s->vars[i] != v) {
t->vars[j] = s->vars[i];
t->exps[j] = s->exps[i];
++j;
continue;
}

/* differentiate with respect to v (this variable) */
t->vars[j] = v;
digit.value[0] = s->exps[i];
if(t->exps[j] = s->exps[i] - 1)
++j; /* variable still present */
t->coef = imul(s->coef, &digit);
++d_flag;
} /* end loop over variables */

if(d_flag == 0) continue; /* v not present, drop term */
if(d_flag > 1) eprint("@variable appears twice in the same term");

t->nvars = j;
++t;
} /* end loop over terms */

if(p->nterms = t - p->terms)  return polysimp(p, 0L);

polyfree(p);
return &pzero;
}

/* take the derivative of a rational function */
struct POLY *deriv(a, v) /*###*/
struct POLY *a;
char v; /* take derivative with respect to v */
{
struct POLY *p, *q, *p1, *p2;

if(!a->next) return pderiv(a, v);

q = pderiv(a->next, v);
if(q == &pzero)
{
p = pderiv(a, v);
p->next = a->next;
return p;
}

p1 = pmul(a, q);
polyfree(q);
q = pderiv(a, v);
p2 = pmul(a->next, q);
polyfree(q);
p = rsub(p2, p1);
bcfree(p1);
bcfree(p2);
p->next = pmul(a->next, a->next);
return p;
}

